/*
 * Copyright (C) 1996-2026 The Squid Software Foundation and contributors
 *
 * Squid software is distributed under GPLv2+ license and includes
 * contributions from numerous individuals and organizations.
 * Please see the COPYING and CONTRIBUTORS files for details.
 */

#ifndef SQUID_SRC_IPC_MEM_FORWARD_H
#define SQUID_SRC_IPC_MEM_FORWARD_H

namespace Ipc
{

namespace Mem
{
typedef uint32_t PoolId;
} // namespace Mem

} // namespace Ipc

#endif /* SQUID_SRC_IPC_MEM_FORWARD_H */

