/*
 * Copyright (C) 1996-2026 The Squid Software Foundation and contributors
 *
 * Squid software is distributed under GPLv2+ license and includes
 * contributions from numerous individuals and organizations.
 * Please see the COPYING and CONTRIBUTORS files for details.
 */
#ifndef SQUID_COMPAT_OS_ANDROID_H
#define SQUID_COMPAT_OS_ANDROID_H

#if defined(__ANDROID__)
/****************************************************************************
 *--------------------------------------------------------------------------*
 * DO *NOT* MAKE ANY CHANGES below here unless you know what you're doing...*
 *--------------------------------------------------------------------------*
 ****************************************************************************/
#define _SQUID_ANDROID_ 1

#endif /* _SQUID_ANDROID_ */
#endif /* SQUID_COMPAT_OS_ANDROID_H */

