/*
 * Copyright (C) 1996-2026 The Squid Software Foundation and contributors
 *
 * Squid software is distributed under GPLv2+ license and includes
 * contributions from numerous individuals and organizations.
 * Please see the COPYING and CONTRIBUTORS files for details.
 */

#ifndef SQUID_SRC_BASE_ONOFF_H
#define SQUID_SRC_BASE_ONOFF_H

/// safer than bool in a list of integer-like function parameters
enum class OnOff { off, on };

#endif /* SQUID_SRC_BASE_ONOFF_H */

