/*
 * Copyright (C) 1996-2026 The Squid Software Foundation and contributors
 *
 * Squid software is distributed under GPLv2+ license and includes
 * contributions from numerous individuals and organizations.
 * Please see the COPYING and CONTRIBUTORS files for details.
 */

#include "squid.h"
#include "internal.h"

#define STUB_API "internal.cc"
#include "tests/STUB.h"

char *internalLocalUri(const char *, const SBuf &) STUB_RETVAL(nullptr)

