/*
 * Copyright (C) 1996-2026 The Squid Software Foundation and contributors
 *
 * Squid software is distributed under GPLv2+ license and includes
 * contributions from numerous individuals and organizations.
 * Please see the COPYING and CONTRIBUTORS files for details.
 */

#ifndef SQUID_SRC_STORE_LOCALSEARCH_H
#define SQUID_SRC_STORE_LOCALSEARCH_H

#include "store/forward.h"

namespace Store {
StoreSearch *NewLocalSearch();
} // namespace Store

#endif /* SQUID_SRC_STORE_LOCALSEARCH_H */

