/*
 * Copyright (C) 1996-2026 The Squid Software Foundation and contributors
 *
 * Squid software is distributed under GPLv2+ license and includes
 * contributions from numerous individuals and organizations.
 * Please see the COPYING and CONTRIBUTORS files for details.
 */

#ifndef SQUID_COMPAT_OS_OS2_H
#define SQUID_COMPAT_OS_OS2_H

#if _SQUID_OS2_

/****************************************************************************
 *--------------------------------------------------------------------------*
 * DO *NOT* MAKE ANY CHANGES below here unless you know what you're doing...*
 *--------------------------------------------------------------------------*
 ****************************************************************************/

/*
*  XXX: the os2 port of bash seems to have problems checking
*  the return codes of programs in if statements.  These options
*  need to be overridden.
*/

#endif /* _SQUID_OS2_ */
#endif /* SQUID_COMPAT_OS_OS2_H */

