/*
 * Copyright (C) 1996-2026 The Squid Software Foundation and contributors
 *
 * Squid software is distributed under GPLv2+ license and includes
 * contributions from numerous individuals and organizations.
 * Please see the COPYING and CONTRIBUTORS files for details.
 */

/* DEBUG: section 18    Cache Manager Statistics */

#include "squid.h"
#include "Store.h"

#define STUB_API "stat.cc"
#include "tests/STUB.h"

class StoreEntry;
const char *storeEntryFlags(const StoreEntry *) STUB_RETVAL(nullptr)

