/*
 * Copyright (C) 1996-2026 The Squid Software Foundation and contributors
 *
 * Squid software is distributed under GPLv2+ license and includes
 * contributions from numerous individuals and organizations.
 * Please see the COPYING and CONTRIBUTORS files for details.
 */

/* DEBUG: section 21    Integer functions */

#ifndef SQUID_SRC_INT_H
#define SQUID_SRC_INT_H

int isPowTen(int); //int.cc

#endif /* SQUID_SRC_INT_H */

