/*
 * Copyright (C) 1996-2026 The Squid Software Foundation and contributors
 *
 * Squid software is distributed under GPLv2+ license and includes
 * contributions from numerous individuals and organizations.
 * Please see the COPYING and CONTRIBUTORS files for details.
 */

/* DEBUG: section 54    Interprocess Communication */

#ifndef SQUID_SRC_IPC_FORWARD_H
#define SQUID_SRC_IPC_FORWARD_H

namespace Ipc
{

class Forwarder;
class Inquirer;
class QuestionerId;
class Request;
class RequestId;
class Response;
class StrandCoord;
class StrandMessage;
class TypedMsgHdr;

} // namespace Ipc

#endif /* SQUID_SRC_IPC_FORWARD_H */

